/* codegen.c
 * R8C (Renesas 16bit microprocessor) Code Generator 
 * Input: Stack-machine Opcodes
 * Output: Opcodes for Registers as Stack cache with Register renaming
 *
 * by Andre Adrian
 *
 * version 28may2006
 */

// uncomment to test stand-alone
// #define MODULTEST

#define REGS 2          // R1, R0

int se = 0;             // Stack elements
int reg[REGS] = {1, 0}; // Register usage

void pushrename() {
  int tmp;
  if (se >= REGS) {
    printf("PUSH.W R%d\n", reg[1]);
  }
  tmp = reg[1];
  reg[1] = reg[0];
  reg[0] = tmp;
  ++se;
}

void poprename() {
  int tmp;
  --se;
  if (se == 0) {
    reg[0] = 1;
    reg[1] = 0;
  } else {
    tmp = reg[0]; 
    reg[0] = reg[1];
    reg[1] = tmp;
  }
  if (se >= REGS) {
    printf("POP.W  R%d\n", reg[1]);
  }
}

pushvar(int ident) {
  printf("\t\t;pushvar %c\n", ident);
  pushrename();
  printf("MOVE.W %c,R%d\n", ident, reg[0]);
}

pushconst(int value) {
  printf("\t\t;pushconst %d\n", value);
  pushrename();
  printf("MOVE.W %d,R%d\n", value, reg[0]);
}

pushaddr(int value) {
  printf("\t\t;push addrof(%c)\n", value);
  printf("MOVA   %c,A0\n", value);
}

pop(int ident) {
  printf("\t\t;pop val, pop addr, *addr=val\n");
  printf("MOVE.W R%d,[A0]\n", reg[0]);
  poprename();
}

add() {
  printf("\t\t;add\n");
  printf("ADD.W  R%d,R%d\n", reg[0], reg[1]);
  poprename();
}

sub() {
  printf("\t\t;sub\n");
  printf("SUB.W  R%d,R%d\n", reg[0], reg[1]);
  poprename();
}

mul() {
  printf("\t\t;mul\n");
  printf("MUL.W  R%d,R%d\n", reg[0], reg[1]);
  poprename();
}  

div() {
  int tmp;
  printf("\t\t;div\n");
  if (reg[0] != 1) {
    printf("XCHG.W R0,R1\n");
    tmp = reg[0]; 
    reg[0] = reg[1];
    reg[1] = tmp;
  }  
  printf("EXTS.W R0\n");
  printf("DIV.W  R1\n");
  poprename();
}  

neg() {
  printf("\t\t;neg\n");
  printf("NEG.W   R%d\n", reg[0]);
}

and() {
  printf("\t\t;and\n");
  printf("AND.W  R%d,R%d\n", reg[0], reg[1]);
  poprename();
}

or() {
  printf("\t\t;or\n");
  printf("OR.W   R%d,R%d\n", reg[0], reg[1]);
  poprename();
}

not() {
  printf("\t\t;not\n");
  printf("NOT.W   R%d\n", reg[0]);
}

#ifdef MODULTEST
main() {
  // c = a - b
  pushvar('a');
  pushvar('b');
  sub();
  pop('c');
  
  pushvar('a');
  pushvar('b');
  pushvar('c');
  pushvar('d');
  pushvar('e');
  pushvar('f');
  sub();
  pop('x');
  sub();
  pop('y');
  sub();
  pop('z');
  
  // a = a - 1
  pushvar('a');
  pushvar('1');
  sub();
  pop('a');
}  
#endif   
