/* symtab.c 
 * Simple symbol table
 *
 * The symbol table is a 1-dim char array. The symbols
 * are separated by space
 * symtab has several sections:
 * keyword section
 * global section for global vars and func names
 * local section for formal parameters and local vars
 *
 * Search with keywords are reserved words
 * first keywords, then locals, then globals
 *
 * by Andre Adrian
 *
 * version 28may2006
 */

#include <string.h>
#include <stdio.h>
 
#define SYMTABLEN 90

char symtab[SYMTABLEN];
int keywords;           /* keywords from index 0 .. keywords */
int globals;            /* globals from index keywords .. globals */
int locals;             /* locals from index globals .. locals */
 
void error(char msg[]) {
  fprintf(stderr, msg);
  fprintf(stderr, "\n");
  exit(1);
}
  
 
int lookup(char name[], int start, int ends) {
// look up name in symbol table. If found return symbol id
// else -1
  char cn, cs;
  int in, is;
  int found, sym;

  in = 0;  cn = name[in];
  is = start;  cs = symtab[is];
  found = 1; sym = 0;
  while (is < ends) {	// symtab end
    if (cn != cs) {
      found = 0;
    }
    is = is + 1; cs = symtab[is];
    if (cn != '\0') {
      in = in + 1; cn = name[in];
    }
    if (cs == ' ') {    // symbol end
      if (found == 1 && cn == '\0') {
      	return sym;
      } else {
	in = 0;  cn = name[in];
	is = is + 1; cs = symtab[is];
	sym = sym + 1;
	found = 1;
      }
    }
  }
  return -1;
}

void addsym(char name[]) {
  if (strlen(symtab) + strlen(name) + 2 <= SYMTABLEN) {
    strcat(symtab, name);
    strcat(symtab, " ");
  } else {
    error("symtab full");
  }
}
 
void initsymtab() {
  strcpy(symtab, "if then else endif while do break wend proc func return int char ");
  keywords = strlen(symtab);
  globals = keywords;
  locals = keywords;
}

int main(int argc, char *argv[]) {
  int sym;
  initsymtab();
  
  sym = lookup("if", 0, keywords);  
  printf("lookup if =%d\n", sym);
  sym = lookup("iff", 0, keywords);  
  printf("lookup iff =%d\n", sym);
  sym = lookup("then", 0, keywords);  
  printf("lookup then =%d\n", sym);
  sym = lookup("char", 0, keywords);  
  printf("lookup char =%d\n", sym);
  sym = lookup("xxx", 0, keywords);  
  printf("lookup xxx =%d\n", sym);
  sym = lookup("yyyyyyyyyyy", 0, keywords);  
  printf("lookup yyyyyyyyyyy =%d\n", sym);
  
  sym = lookup("array", 0, keywords);  
  printf("lookup array =%d\n", sym);
  addsym("array");
  sym = lookup("array", 0, keywords);  
  printf("lookup array =%d\n", sym);
  
  addsym("yyyyyyyyyyy");
  sym = lookup("yyyyyyyyyyy", 0, keywords);  
  printf("lookup yyyyyyyyyyy =%d\n", sym);
  
  addsym("zzzzzzzzzzz");
  sym = lookup("zzzzzzzzzzz", 0, keywords);  
  printf("lookup zzzzzzzzzzz =%d\n", sym);  
}
