%{
  /* tbng01.y
   * Tiny BASIC no GOTO parser
   *
   * by Andre Adrian
   *
   * version 28may2006
   */ 
  #include <stdio.h>
  #include "codegen.h"
  int yylex(void);
  void yyerror(char *);

  char func;
%}

%token  CONST CHRCONST STRCONST IDENT
%token	VAR INT CHAR ARRAY VOID
%token	FUNC FORWARD BEGINN END RETURN
%token	IF THEN ELSE ENDIF BREAK
%token	WHILE DO WEND
%token  LE GE

%%

/* based on Pascal BNF */

program:  
   block4  
   | variable_declaration sc block4
   ;
   
sc:
   ';'
   | sc ';'   
   ;
   
block4:  
   func_declaration sc
   ;
   
variable_declaration:  
   VAR global_decl  
   | variable_declaration sc global_decl
   ;
   
global_decl:
   IDENT ':' INT                 { printf("global %c int\n", $1); }                      
   | IDENT ':' CHAR              { printf("global %c char\n", $1); }                     
   | IDENT ':' ARRAY CONST INT   { printf("global %c int_array %d\n", $1, $4); }         
   | IDENT ':' ARRAY CONST CHAR  { printf("global %c char_array %d\n", $1, $4); }
   ;

func_declaration:
   func
   | func_declaration sc func
   ;

func:
   funchead '(' param_list ')' ':' return_type sc block_or_forward 
   ;

funchead:
   FUNC IDENT                    { printf("func %c decl\n", $2); func = $2;}
   ;

param_list:
   IDENT ':' param_type
   | param_list ',' IDENT ':' param_type
   |
   ;

param_type:
   INT                           { printf("param %c int\n", $1); }        
   | CHAR                        { printf("param %c char\n", $1); }       
   | ARRAY INT                   { printf("param %c int_array\n", $1); }  
   | ARRAY CHAR                  { printf("param %c char_array\n", $1); } 
   ;
      
return_type:
   INT                           { printf("func %c returns int\n", func); }        
   | CHAR                        { printf("func %c returns char\n", func); }       
   | VOID                        { printf("func %c returns void\n", func); }       
   ;
   
block_or_forward:  
   pblock3                       
   | FORWARD                     { printf("forward decl\n"); }            
   ;
   
pblock3:  
   pblock4  
   | pvariable_declaration sc pblock4
   ;
   
pblock4:  
   begin_label stmt_list END  
   ;
   
begin_label:
   BEGINN                        { printf("func_label %c\n", func); }
   ;
   
pvariable_declaration:  
   VAR local_decl  
   | pvariable_declaration sc local_decl
   ;
   
local_decl:
   IDENT ':' INT                 { printf("local %c int\n", $1); }                       
   | IDENT ':' CHAR              { printf("local %c char\n", $1); }                      
   | IDENT ':' ARRAY CONST INT   { printf("local %c int_array %d\n", $1, $4); }          
   | IDENT ':' ARRAY CONST CHAR  { printf("local %c char_array %d\n", $1, $4); }         
   ;

stmt_list:  
   statement ';'
   | stmt_list statement ';'
   ;  

statement:
   variable '=' expr             { pop(); }
   | IF cond THEN stmt_list else_or_endif
   | WHILE cond DO stmt_list WEND 
   | IDENT '(' expr_list ')'     { printf("call %c(), drop args\n", $1); }
   | RETURN                      { printf("return\n"); }
   | RETURN expr                 { printf("pop to return-reg, return\n"); }
   | BREAK 
   | 
   ;

else_or_endif:
   ENDIF                            { printf("endif_label\n"); }
   | else_label stmt_list ENDIF     { printf("endif_label\n"); }
   ;
   
else_label:
   ELSE                      { printf("jump endif_label, else_label\n"); }   
   
variable:
   IDENT                     { pushaddr($1); }                 
   | IDENT '[' expr ']'      { printf("pop ndx, push addrof(%c[ndx])\n", $1); }   
   ;  
    
expr_list:
   expr                      { printf("incr arg cnt\n", $1); }                    
   | expr_list ',' expr      { printf("incr arg cnt\n", $1); }                    
   |
   ;
  
cond:
   expr '=' expr       	     { printf("pop r, pop l, branch if l != r\n"); }                           
   | expr '#' expr           { printf("pop r, pop l, branch if l == r\n"); }
   | expr '<' expr           { printf("pop r, pop l, branch if l >= r\n"); }
   | expr '>' expr           { printf("pop r, pop l, branch if l <= r\n"); }
   | expr LE  expr           { printf("pop r, pop l, branch if l > r\n"); }
   | expr GE  expr           { printf("pop r, pop l, branch if l < r\n"); }
   ;

expr:
   term
   | expr '+' term           { add();}
   | expr '-' term           { sub(); }
   | expr '|' term           { or(); }
   ;

term:
   unary_fact
   | term '*' unary_fact     { mul(); }                      
   | term '/' unary_fact     { div(); }                      
   | term '&' unary_fact     { and(); }                      
   ;

unary_fact: 
   '-' unary_fact            { neg(); }                          
   | '!' unary_fact          { not(); }                          
   | fact
   ;

fact:
   CONST      	      	     { pushconst($1); }                     
   | CHRCONST      	     { printf("push char const %c\n", $1); }                     
   | STRCONST      	     { printf("push array char const %c..\n", $1); }                     
   | IDENT                   { pushvar($1); }                       
   | IDENT '[' expr ']'      { printf("pop ndx, push %c[ndx]\n", $1); }            
   | '(' expr ')'
   | IDENT '(' expr_list ')' { printf("call %c(), drop args, push return-reg\n", $1); }             
   ;

%%

void yyerror(char *s) {
    fprintf(stderr, "%s\n", s);
}

int main(void) {
#if YYDEBUG
    yydebug = 1;
#endif    
    yyparse();
    return 0;
}
